# file generated by setuptools_scm
# don't change, don't track in version control
TYPE_CHECKING = False
if TYPE_CHECKING:
    from typing import Tuple

__version__ = version = '1.3.1'  # type: str
__version_tuple__ = version_tuple = (1, 3, 1)  # type: Tuple[int | str, ...]
